#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

#include "config.h"
#include "imapfilter.h"


extern conn_t connpri, connaux;
extern unsigned int options;


int action_delete(char *mesgs, char *args);
int action_copy(char *mbox, char *mesgs, char *destmbox, char *args);
int action_move(char *mbox, char *mesgs, char *destmbox, char *args);
int action_rcopy(char *mbox, char *mesgs, account_t * destacc, char *destmbox, char *args);
int action_rmove(char *mbox, char *mesgs, account_t * destacc, char *destmbox, char *args);
int action_flag(char *mesgs, unsigned int *type, unsigned int *msgflags, char *args);
int action_list(char *mesgs, char *args);

unsigned int count_messages(char *mesgs);
char *convert_messages(char *mesgs);
int substitute_date(char *str);
void current_date(char *destmbox);
void message_date(char *mesg, char *destmbox);
void default_variables(char *mbox, char *destmbox);


/*
 * Apply the appropriate action.
 */
int
apply_action(char *mbox, char *mesgs, unsigned int *type, account_t * raccount,
    char *destmbox, unsigned int *msgflags, char *args)
{
	unsigned int cnt;

	if (*mesgs == '\0')
		return 0;

	log_info(LOG_ACTION, type);
	log_info(LOG_DESTINATION_ACCOUNT, raccount->key);
	log_info(LOG_DESTINATION_MAILBOX, destmbox);

	cnt = count_messages(mesgs);

	switch (*type) {
	case FILTER_ACTION_DELETE:
		info("%d message%s deleted.\n", cnt, plural(cnt));
		action_delete(mesgs, args);
		break;
	case FILTER_ACTION_COPY:
		info("%d message%s copied from \"%s\" to mailbox \"%s\".\n",
		    cnt, plural(cnt), mbox, destmbox);
		action_copy(mbox, mesgs, apply_namespace(destmbox,
			connpri.nsp.prefix, connpri.nsp.delim), args);
		break;
	case FILTER_ACTION_MOVE:
		info("%d message%s moved from \"%s\" to mailbox \"%s\".\n",
		    cnt, plural(cnt), mbox, destmbox);
		action_move(mbox, mesgs, apply_namespace(destmbox,
			connpri.nsp.prefix, connpri.nsp.delim), args);
		break;
	case FILTER_ACTION_RCOPY:
		info("%d message%s copied from \"%s\" to mailbox "
		    "\"%s\" at account %s.\n", cnt, plural(cnt),
		    mbox, destmbox, raccount->key);
		action_rcopy(mbox, mesgs, raccount, destmbox, args);
		break;
	case FILTER_ACTION_RMOVE:
		info("%d message%s moved from \"%s\" to mailbox "
		    "\"%s\" at account %s.\n", cnt, plural(cnt),
		    mbox, destmbox, raccount->key);
		action_rmove(mbox, mesgs, raccount, destmbox, args);
		break;
	case FILTER_ACTION_FLAG_REPLACE:
	case FILTER_ACTION_FLAG_ADD:
	case FILTER_ACTION_FLAG_REMOVE:
		info("%d message%s flagged.\n", cnt, plural(cnt));
		action_flag(mesgs, type, msgflags, args);
		break;
	case FILTER_ACTION_LIST:
		info("%d message%s listed.\n", cnt, plural(cnt));
		action_list(mesgs, args);
		break;
	}

	if (*args == '\0')
		log_info(LOG_PREAMBLE, NULL);

	return 0;
}


/*
 * Delete messages and optionally list some of their headers.
 */
int
action_delete(char *mesgs, char *args)
{
	char *tok, *m, *mcp;

	action_list(mesgs, args);

	m = mcp = convert_messages(mesgs);

	tok = strtok_r(m, " ", &m);
	while (tok) {
		server_response(&connpri, imap_store(&connpri, tok,
			STORE_FLAG_ADD, "\\Deleted"));

		tok = strtok_r(NULL, " ", &m);
	}

	if (options & OPTION_EXPUNGE)
		server_response(&connpri, imap_expunge(&connpri));

	xfree(mcp);

	return 0;
}


/*
 * Copy messages to specified mailbox.
 */
int
action_copy(char *mbox, char *mesgs, char *destmbox, char *args)
{
	int r;
	char *tok, *mcp, *m;
	char dm[2][MBOX_NAME_LEN];

	r = 0;
	tok = NULL;

	action_list(mesgs, args);

	if (strchr(destmbox, '@'))
		m = mcp = xstrdup(mesgs);
	else
		m = mcp = convert_messages(mesgs);

	xstrncpy(dm[0], destmbox, MBOX_NAME_LEN - 1);
	default_variables(mbox, dm[0]);
	current_date(dm[0]);
	tok = strtok_r(m, " ", &m);
	while (tok != NULL) {
		xstrncpy(dm[1], dm[0], MBOX_NAME_LEN - 1);
		message_date(tok, dm[1]);

		if ((r = copy_response(&connpri, imap_copy(&connpri, tok,
				dm[1]))) == RESPONSE_TRYCREATE)
			if (!server_response(&connpri, imap_create(&connpri,
				    dm[1]))) {
				if ((options & OPTION_SUBSCRIBE))
					server_response(&connpri,
					    imap_subscribe(&connpri, dm[1]));
				r = copy_response(&connpri,
				    imap_copy(&connpri, tok, dm[1]));
			}
		tok = strtok_r(NULL, " ", &m);
	}

	xfree(mcp);

	return r;
}


/*
 * Move messages to specified mailbox.
 */
int
action_move(char *mbox, char *mesgs, char *destmbox, char *args)
{
	if (!action_copy(mbox, mesgs, destmbox, args))
		action_delete(mesgs, "\0");

	return 0;
}


/*
 * Copy messages to the specified mailbox of another mail server.
 */
int
action_rcopy(char *mbox, char *mesgs, account_t * destacc, char *destmbox,
    char *args)
{
	int r, ta, tf;
	char *tok, *m, *mcp, *ndm;
	unsigned int n;
	char buf[RESPONSE_BUF * 2 + 1];
	char dm[3][MBOX_NAME_LEN];

	*dm[0] = *dm[1] = *dm[2] = '\0';

	if (init_connection(&connaux, destacc->server, destacc->port,
		destacc->ssl))
		return ERROR_NETWORK;

	r = greeting_response(&connaux);

#ifdef DEBUG
	test(&connaux);
#endif

	if (r == RESPONSE_BYE || check_capabilities(&connaux))
		return ERROR_NETWORK;

#ifdef SSL_TLS
	if (destacc->ssl == SSL_DISABLED && connaux.caps & CAPABILITY_STARTTLS)
		if (negotiate_tls(&connaux) == RESPONSE_OK)
			check_capabilities(&connaux);
#endif

	if (r != RESPONSE_PREAUTH) {
		if (destacc->passwdattr == PASSWORD_NONE) {
			printf("Enter password for %s@%s: ", destacc->username,
			    destacc->server);
			get_password(destacc->password, PASSWORD_LEN);
			destacc->passwdattr = PASSWORD_PLAIN;
		}
#ifdef CRAM_MD5
		if (connaux.caps & CAPABILITY_AUTH_CRAM_MD5)
			r = auth_cram_md5(&connaux, destacc->username,
			    destacc->password);
		else
#endif
			r = login(&connaux, destacc->username,
			    destacc->password);

		if (r == RESPONSE_NO) {
			error("username %s or password rejected at %s\n",
			    destacc->username, destacc->server);
			return ERROR_NETWORK;
		}
	}
	check_namespace(&connaux);

	m = mcp = xstrdup(mesgs);

	xstrncpy(dm[1], destmbox, MBOX_NAME_LEN - 1);
	default_variables(mbox, dm[1]);
	current_date(dm[1]);

	tok = strtok_r(m, " ", &m);
	while (tok != NULL) {
		xstrncpy(dm[2], dm[1], MBOX_NAME_LEN - 1);
		message_date(tok, dm[2]);

		/* apply_namespace() returns a pointer to a static buffer. */
		ndm = apply_namespace(dm[2], connaux.nsp.prefix,
		    connaux.nsp.delim);

		/* Check only if mailbox name is different from last one. */
		if (strncmp(dm[0], dm[2], strlen(dm[2]))) {
			r = check_mailbox(&connaux, ndm);
			if (r == RESPONSE_NO) {
				server_response(&connaux,
				    imap_create(&connaux, ndm));
				if ((options & OPTION_SUBSCRIBE))
					server_response(&connaux,
					    imap_subscribe(&connaux, ndm));
			}
		}
		xstrncpy(dm[0], dm[2], MBOX_NAME_LEN - 1);

		fetchsize_response(&connpri, &n,
		    imap_fetch(&connpri, tok, "RFC822.SIZE"));

		ta = imap_append(&connaux, ndm, n);

		fetch_response(&connpri, 0, 1, NULL);
		tf = imap_fetch(&connpri, tok, "RFC822.HEADER");
		do {
			r = fetch_response(&connpri, tf, 0, buf);
			socket_write(&connaux, buf);
		} while (r == RESPONSE_NONE);

		socket_write(&connaux, "\r\n");

		fetch_response(&connpri, 0, 1, NULL);
		tf = imap_fetch(&connpri, tok, "BODY[TEXT]");
		do {
			r = fetch_response(&connpri, tf, 0, buf);
			if (r != RESPONSE_NULLBODY)
				socket_write(&connaux, buf);
		} while (r == RESPONSE_NONE);

		if (r != RESPONSE_NULLBODY)
			socket_write(&connaux, "\r\n\r\n");
		else
			socket_write(&connaux, "\r\n");

		append_response(&connaux, ta);

		tok = strtok_r(NULL, " ", &m);
	}

	logout(&connaux);

	action_list(mesgs, args);

	xfree(mcp);

	return 0;
}


/*
 * Move messages to the specified mailbox of another mail server.
 */
int
action_rmove(char *mbox, char *mesgs, account_t * destacc, char *destmbox,
    char *args)
{
	if (!action_rcopy(mbox, mesgs, destacc, destmbox, args))
		action_delete(mesgs, "\0");

	return 0;
}


/*
 * Flag messages by replacing, adding or removing specified flags.
 */
int
action_flag(char *mesgs, unsigned int *type, unsigned int *msgflags,
    char *args)
{
	unsigned int t;
	char s[STORE_FLAGS_BUF];
	char *tok, *m, *mcp;

	*s = 0;

	switch (*type) {
	case FILTER_ACTION_FLAG_ADD:
		t = STORE_FLAG_ADD;
		break;
	case FILTER_ACTION_FLAG_REMOVE:
		t = STORE_FLAG_REMOVE;
		break;
	default:
		t = STORE_FLAG_REPLACE;
	}

	if ((*msgflags != MESSAGE_FLAG_NONE)) {
		if ((*msgflags & MESSAGE_FLAG_SEEN))
			strncat(s, "\\Seen ", STORE_FLAGS_BUF - strlen(s) - 1);
		if ((*msgflags & MESSAGE_FLAG_ANSWERED))
			strncat(s, "\\Answered ",
			    STORE_FLAGS_BUF - strlen(s) - 1);
		if ((*msgflags & MESSAGE_FLAG_FLAGGED))
			strncat(s, "\\Flagged ",
			    STORE_FLAGS_BUF - strlen(s) - 1);
		if ((*msgflags & MESSAGE_FLAG_DELETED))
			strncat(s, "\\Deleted ",
			    STORE_FLAGS_BUF - strlen(s) - 1);
		if ((*msgflags & MESSAGE_FLAG_DRAFT))
			strncat(s, "\\Draft", STORE_FLAGS_BUF - strlen(s) - 1);
		if ((s[strlen(s) - 1] == ' '))
			s[strlen(s) - 1] = 0;
	}
	action_list(mesgs, args);

	m = mcp = convert_messages(mesgs);

	tok = strtok_r(m, " ", &m);
	while (tok != NULL) {
		server_response(&connpri, imap_store(&connpri, tok, t, s));

		tok = strtok_r(NULL, " ", &m);
	}

	if (options & OPTION_EXPUNGE)
		server_response(&connpri, imap_expunge(&connpri));

	xfree(mcp);

	return 0;
}

/*
 * List user selected headers of messages.
 */
int
action_list(char *mesgs, char *args)
{
	int r, t;
	char *tok, *mcp, *m;
	char s[ARGS_LEN + 27];
	char hdrs[RESPONSE_BUF * 2 + 1];

	if (*args == '\0')
		return 0;

	m = mcp = xstrdup(mesgs);

	snprintf(s, ARGS_LEN + 27 - 1, "BODY.PEEK[HEADER.FIELDS (%s)]", args);

	tok = strtok_r(m, " ", &m);
	while (tok != NULL) {
		/* Reset internal fetch counter. */
		fetch_response(&connpri, 0, 1, NULL);
		t = imap_fetch(&connpri, tok, s);

		log_info(LOG_PREAMBLE, NULL);
		do {
			r = fetch_response(&connpri, t, 0, hdrs);

			if (*hdrs != '\0') {
				if (options & OPTION_HEADERS)
					info("%s\n", hdrs);
				log_info(LOG_HEADER, hdrs);
			}
		} while (r == RESPONSE_NONE);

		tok = strtok_r(NULL, " ", &m);
	}

	xfree(mcp);

	return 0;
}


/*
 * Count how many messages matched the filter.
 */
unsigned int
count_messages(char *mesgs)
{
	unsigned int cnt;
	char *c;

	cnt = 0;
	c = mesgs;

	while ((c = strchr(c, ' '))) {
		cnt++;
		c++;
	}

	return ++cnt;
}


/*
 * Convert messages with contiguous sequence number to the corresponding
 * number range, eg. 1 2 3 5 7 8 --> 1:3 5 7:8 and return a newly allocated
 * buffer with the results.
 */
char *
convert_messages(char *mesgs)
{
	int maxlen;
	unsigned int start, end, tmp;
	char *c, *cp, *tail;

	start = end = tmp = 0;
	maxlen = strlen(mesgs) + 1;
	tail = NULL;

	c = cp = xstrdup(mesgs);

	start = (unsigned int)strtoul(mesgs, &tail, 10);
	end = start;

	do {
		if (tail != NULL) {
			tmp = (unsigned int)strtoul(tail, &tail, 10);
			if (tmp == 0)
				tail = NULL;
		}
		if (tmp == end + 1)
			end++;
		else {
			if (start == end) {
				xstrncpy(c, ultostr(start, 10), maxlen);
				c += strlen(c);
			} else {
				xstrncpy(c, ultostr(start, 10), maxlen - 1);
				c += strlen(c);
				*c = ':';
				*++c = 0;
				xstrncpy(c, ultostr(end, 10), maxlen);
				c += strlen(c);
			}

			if (tail != NULL && c - cp < maxlen) {
				*c = ' ';
				*++c = 0;
			}
			start = end = tmp;
		}
	} while (tmp != 0);

	return cp;
}


/*
 * Substitute all occurences of the '@' character with the '%' character,
 * and any double '@' characters with a signle '@' character, returning the
 * number of replacements.
 */
int
substitute_date(char *str)
{
	int n;
	char *r, *w, *s;

	n = 0;

	s = xstrdup(str);

	for (r = s, w = str; *r != '\0'; r++, w++) {
		if (*r == '%') {
			*w++ = '%';
			*w = '%';
		} else if (*r == '@') {
			if (*(r + 1) == '@') {
				*w = *r++;
			} else {
				*w = '%';
				n++;
			}
		} else {
			*w = *r;
		}
	}
	*w = '\0';

	xfree(s);

	return n;
}


/*
 * Format the destination mailbox according to the current date conversion
 * specifiers.
 */
void
current_date(char *destmbox)
{
	char s[MBOX_NAME_LEN];
	time_t te;
	struct tm *tl;

	if (!strchr(destmbox, '%'))
		return;

	te = time(NULL);
	tl = localtime(&te);

	if (strftime(s, MBOX_NAME_LEN - 1, destmbox, tl))
		xstrncpy(destmbox, s, MBOX_NAME_LEN - 1);
}


/*
 * Format the destination mailbox according to the message date conversion
 * specifiers.
 */
void
message_date(char *mesg, char *destmbox)
{
	unsigned int t;
	char s[MBOX_NAME_LEN];
	struct tm tl;
	char dbuf[RESPONSE_BUF + 1];

	if (!strchr(destmbox, '@'))
		return;

	substitute_date(destmbox);

	fetch_response(&connpri, 0, 1, NULL);
	t = imap_fetch(&connpri, mesg, "BODY.PEEK[HEADER.FIELDS (DATE)]");

	while (fetch_response(&connpri, t, 0, dbuf) == RESPONSE_NONE);

	if (strptime(dbuf, "Date: %a, %d %b %Y %H:%M:%S", &tl) &&
	    strftime(s, MBOX_NAME_LEN - 1, destmbox, &tl))
		xstrncpy(destmbox, s, MBOX_NAME_LEN - 1);
}


/*
 * Format the destination mailbox according to "default variables" specifiers.
 */
void
default_variables(char *mbox, char *destmbox)
{
	char *m, *r, *w, *s;

	if (!strchr(destmbox, '$'))
		return;

	s = xstrdup(destmbox);

	for (r = s, w = destmbox; *r != '\0';) {
		if (w - destmbox >= MBOX_NAME_LEN - 1)
			break;
		if (*r == '$') {
			switch (*(r + 1)) {
			case '_':
				if (w + strlen(mbox) - destmbox >
				    MBOX_NAME_LEN - 1) {
					r += 2;
					break;
				}
				for (m = mbox; *m != '\0'; m++, w++)
					*w = *m;
				r += 2;
				break;
			case '$':
				*w++ = '$';
				r += 2;
				break;
			default:
				*w++ = *r++;
				break;
			}
		} else {
			*w++ = *r++;
		}
	}
	*w = '\0';

	xfree(s);
}
